@@ -9,7 +9,7 @@ use bevy::{
         system::{lifetimeless::*, SystemState},
     },
     log::trace,
-    math::{const_vec3, Mat4, Vec2, Vec3, Vec4Swizzles},
+    math::{const_vec3, Mat4, Vec2, Vec3, Vec4, Vec4Swizzles},
     reflect::TypeUuid,
     render::{
         color::Color,
@@ -33,7 +33,12 @@ use rand::Rng;
 use std::sync::atomic::{AtomicU64, Ordering as AtomicOrdering};
 use std::{borrow::Cow, cmp::Ordering, num::NonZeroU64, ops::Range};
 
-use crate::{asset::EffectAsset, Gradient, ParticleEffect, ToWgslFloat};
+use crate::{
+    asset::EffectAsset,
+    modifiers::{ForceFieldParam, FFNUM},
+    spawn::{new_rng, Random},
+    Gradient, ParticleEffect, ToWgslString,
+};
 
 mod compute_cache;
 mod effect_cache;
@@ -60,6 +65,13 @@ const DEFAULT_POSITION_CODE: &'static str = r##"
     ret.vel = dir * speed;
 "##;
 
+const DEFAULT_FORCE_FIELD_CODE: &'static str = r##"
+    vVel = vVel + (spawner.accel * sim_params.dt);
+    vPos = vPos + vVel * sim_params.dt;
+"##;
+
+const FORCE_FIELD_CODE: &'static str = include_str!("force_field_code.wgsl");
+
 /// Labels for the Hanabi systems.
 #[derive(Debug, Hash, PartialEq, Eq, Clone, SystemLabel)]
 pub enum EffectSystems {
@@ -73,11 +85,13 @@ pub enum EffectSystems {
     QueueEffects,
 }
 
+/// Trait to convert any data structure to its equivalent shader code.
 trait ShaderCode {
+    /// Generate the shader code for the current state of the object.
     fn to_shader_code(&self) -> String;
 }
 
-impl ShaderCode for Gradient {
+impl ShaderCode for Gradient<Vec2> {
     fn to_shader_code(&self) -> String {
         if self.keys().len() == 0 {
             return String::new();
@@ -88,19 +102,59 @@ impl ShaderCode for Gradient {
             .enumerate()
             .map(|(index, key)| {
                 format!(
-                    "let t{0} = {1};\nlet c{0} = vec4<f32>({2}, {3}, {4}, {5});",
+                    "let t{0} = {1};\nlet v{0} = {2};",
                     index,
-                    key.ratio.to_float_string(),
-                    key.color.x.to_float_string(),
-                    key.color.y.to_float_string(),
-                    key.color.z.to_float_string(),
-                    key.color.w.to_float_string()
+                    key.ratio().to_wgsl_string(),
+                    key.value.to_wgsl_string()
+                )
+            })
+            .fold("// Gradient\n".into(), |s, key| s + &key + "\n");
+        if self.keys().len() == 1 {
+            s + "size = v0;\n"
+        } else {
+            // FIXME - particle.age and particle.lifetime are unrelated to Gradient<Vec4>
+            s += "let life = particle.age / particle.lifetime;\nif (life <= t0) { size = v0; }\n";
+            let mut s = self
+                .keys()
+                .iter()
+                .skip(1)
+                .enumerate()
+                .map(|(index, _key)| {
+                    format!(
+                        "else if (life <= t{1}) {{ size = mix(v{0}, v{1}, (life - t{0}) / (t{1} - t{0})); }}\n",
+                        index,
+                        index + 1
+                    )
+                })
+                .fold(s, |s, key| s + &key);
+            s += &format!("else {{ size = v{}; }}\n", self.keys().len() - 1);
+            s
+        }
+    }
+}
+
+impl ShaderCode for Gradient<Vec4> {
+    fn to_shader_code(&self) -> String {
+        if self.keys().len() == 0 {
+            return String::new();
+        }
+        let mut s: String = self
+            .keys()
+            .iter()
+            .enumerate()
+            .map(|(index, key)| {
+                format!(
+                    "let t{0} = {1};\nlet c{0} = {2};",
+                    index,
+                    key.ratio().to_wgsl_string(),
+                    key.value.to_wgsl_string()
                 )
             })
             .fold("// Gradient\n".into(), |s, key| s + &key + "\n");
         if self.keys().len() == 1 {
             s + "out.color = c0;\n"
         } else {
+            // FIXME - particle.age and particle.lifetime are unrelated to Gradient<Vec4>
             s += "let life = particle.age / particle.lifetime;\nif (life <= t0) { out.color = c0; }\n";
             let mut s = self
                 .keys()
@@ -156,6 +210,30 @@ impl From<SimParams> for SimParamsUniform {
     }
 }
 
+#[repr(C)]
+#[derive(Debug, Default, Clone, Copy, Pod, Zeroable, AsStd430)]
+pub struct ForceFieldStd430 {
+    pub position_or_direction: Vec3,
+    pub max_radius: f32,
+    pub min_radius: f32,
+    pub mass: f32,
+    pub force_exponent: f32,
+    pub conform_to_sphere: f32,
+}
+
+impl Into<ForceFieldStd430> for ForceFieldParam {
+    fn into(self) -> ForceFieldStd430 {
+        ForceFieldStd430 {
+            position_or_direction: self.position,
+            max_radius: self.max_radius,
+            min_radius: self.min_radius,
+            mass: self.mass,
+            force_exponent: self.force_exponent,
+            conform_to_sphere: if self.conform_to_sphere { 1.0 } else { 0.0 },
+        }
+    }
+}
+
 #[repr(C)]
 #[derive(Debug, Default, Copy, Clone, Pod, Zeroable, AsStd430)]
 struct SpawnerParams {
@@ -167,8 +245,12 @@ struct SpawnerParams {
     /// Global acceleration applied to all particles each frame.
     /// TODO - This is NOT a spawner/emitter thing, but is a per-effect one. Rename SpawnerParams?
     accel: Vec3,
+
     /// Current number of used particles.
     count: i32,
+
+    /// Force field components. One PullingForceFieldParam takes up 32 bytes.
+    force_field: [ForceFieldStd430; FFNUM],
     ///
     __pad0: Vec3,
     /// Spawn seed, for randomized modifiers.
@@ -183,6 +265,7 @@ pub struct ParticlesUpdatePipeline {
     sim_params_layout: BindGroupLayout,
     particles_buffer_layout: BindGroupLayout,
     spawner_buffer_layout: BindGroupLayout,
+    indirect_buffer_layout: BindGroupLayout,
     pipeline_layout: PipelineLayout,
 }
 
@@ -191,6 +274,12 @@ impl FromWorld for ParticlesUpdatePipeline {
         let world = world.cell();
         let render_device = world.get_resource::<RenderDevice>().unwrap();
 
+        let limits = render_device.limits();
+        bevy::log::info!(
+            "GPU limits:\n- max_compute_invocations_per_workgroup={}\n- max_compute_workgroup_size_x={}\n- max_compute_workgroup_size_y={}\n- max_compute_workgroup_size_z={}\n- max_compute_workgroups_per_dimension={}",
+            limits.max_compute_invocations_per_workgroup, limits.max_compute_workgroup_size_x, limits.max_compute_workgroup_size_y, limits.max_compute_workgroup_size_z, limits.max_compute_workgroups_per_dimension
+        );
+
         trace!(
             "SimParamsUniform: std140_size_static = {}",
             SimParamsUniform::std140_size_static()
@@ -252,12 +341,28 @@ impl FromWorld for ParticlesUpdatePipeline {
                 label: Some("particles_update_spawner_buffer_layout"),
             });
 
+        let indirect_buffer_layout =
+            render_device.create_bind_group_layout(&BindGroupLayoutDescriptor {
+                entries: &[BindGroupLayoutEntry {
+                    binding: 0,
+                    visibility: ShaderStages::COMPUTE,
+                    ty: BindingType::Buffer {
+                        ty: BufferBindingType::Storage { read_only: false },
+                        has_dynamic_offset: true,
+                        min_binding_size: BufferSize::new(std::mem::size_of::<u32>() as u64),
+                    },
+                    count: None,
+                }],
+                label: Some("particles_update_indirect_buffer_layout"),
+            });
+
         let pipeline_layout = render_device.create_pipeline_layout(&PipelineLayoutDescriptor {
             label: Some("particles_update_pipeline_layout"),
             bind_group_layouts: &[
                 &sim_params_layout,
                 &particles_buffer_layout,
                 &spawner_buffer_layout,
+                &indirect_buffer_layout,
             ],
             push_constant_ranges: &[],
         });
@@ -266,6 +371,7 @@ impl FromWorld for ParticlesUpdatePipeline {
             sim_params_layout,
             particles_buffer_layout,
             spawner_buffer_layout,
+            indirect_buffer_layout,
             pipeline_layout,
         }
     }
@@ -345,12 +451,14 @@ impl FromWorld for ParticlesRenderPipeline {
 pub struct ParticleUpdatePipelineKey {
     /// Code for the position initialization of newly emitted particles.
     position_code: String,
+    force_field_code: String,
 }
 
 impl Default for ParticleUpdatePipelineKey {
     fn default() -> Self {
         ParticleUpdatePipelineKey {
             position_code: Default::default(),
+            force_field_code: Default::default(),
         }
     }
 }
@@ -359,8 +467,11 @@ impl SpecializedComputePipeline for ParticlesUpdatePipeline {
     type Key = ParticleUpdatePipelineKey;
 
     fn specialize(&self, key: Self::Key, render_device: &RenderDevice) -> ComputePipeline {
-        let source =
+        let mut source =
             PARTICLES_UPDATE_SHADER_TEMPLATE.replace("{{INIT_POS_VEL}}", &key.position_code);
+
+        source = source.replace("{{FORCE_FIELD_CODE}}", &key.force_field_code);
+
         //trace!("Specialized compute pipeline:\n{}", source);
 
         let shader_module = render_device.create_shader_module(&ShaderModuleDescriptor {
@@ -489,7 +600,8 @@ impl SpecializedPipeline for ParticlesRenderPipeline {
             depth_stencil: Some(DepthStencilState {
                 format: TextureFormat::Depth32Float,
                 depth_write_enabled: false,
-                depth_compare: CompareFunction::Always,
+                // Bevy uses reverse-Z, so Greater really means closer
+                depth_compare: CompareFunction::Greater,
                 stencil: StencilState::default(),
                 bias: DepthBiasState::default(),
             }),
@@ -517,6 +629,8 @@ pub struct ExtractedEffect {
     pub transform: Mat4,
     /// Constant acceleration applied to all particles.
     pub accel: Vec3,
+    /// Force field applied to all particles in the "update" phase.
+    force_field: [ForceFieldParam; FFNUM],
     /// Particles tint to modulate with the texture image.
     pub color: Color,
     pub rect: Rect,
@@ -529,11 +643,13 @@ pub struct ExtractedEffect {
     pub shader: Handle<Shader>,
     /// Update position code.
     pub position_code: String,
+    /// Update force field code.
+    pub force_field_code: String,
 }
 
-/// Extracted data for newly-added ParticleEffect component requiring a new GPU allocation.
+/// Extracted data for newly-added [`ParticleEffect`] component requiring a new GPU allocation.
 pub struct AddedEffect {
-    /// Entity with a newly-added ParticleEffect component.
+    /// Entity with a newly-added [`ParticleEffect`] component.
     pub entity: Entity,
     /// Capacity of the effect (and therefore, the particle buffer), in number of particles.
     pub capacity: u32,
@@ -549,7 +665,7 @@ pub struct AddedEffect {
 pub struct ExtractedEffects {
     /// Map of extracted effects from the entity the source [`ParticleEffect`] is on.
     pub effects: HashMap<Entity, ExtractedEffect>,
-    /// Entites which had their ParticleEffect component removed.
+    /// Entites which had their [`ParticleEffect`] component removed.
     pub removed_effect_entities: Vec<Entity>,
     /// Newly added effects without a GPU allocation yet.
     pub added_effects: Vec<AddedEffect>,
@@ -599,6 +715,7 @@ pub(crate) fn extract_effects(
     _images: Res<Assets<Image>>,
     mut shaders: ResMut<Assets<Shader>>,
     mut pipeline_registry: ResMut<PipelineRegistry>,
+    mut rng: ResMut<Random>,
     mut query: QuerySet<(
         // All existing ParticleEffect components
         QueryState<(
@@ -661,10 +778,12 @@ pub(crate) fn extract_effects(
 
             // Tick the effect's spawner to determine the spawn count for this frame
             let spawner = effect.spawner(&asset.spawner);
-            let spawn_count = spawner.tick(dt);
+
+            let spawn_count = spawner.tick(dt, &mut rng.0);
 
             // Extract the acceleration
             let accel = asset.update_layout.accel;
+            let force_field = asset.update_layout.force_field;
 
             // Generate the shader code for the position initializing of newly emitted particles
             // TODO - Move that to a pre-pass, not each frame!
@@ -675,14 +794,28 @@ pub(crate) fn extract_effects(
                 position_code.clone()
             };
 
+            // Generate the shader code for the force field of newly emitted particles
+            // TODO - Move that to a pre-pass, not each frame!
+            // let force_field_code = &asset.init_layout.force_field_code;
+            // let force_field_code = if force_field_code.is_empty() {
+            let force_field_code = if 0.0 == asset.update_layout.force_field[0].force_exponent {
+                DEFAULT_FORCE_FIELD_CODE.to_owned()
+            } else {
+                FORCE_FIELD_CODE.to_owned()
+            };
+
             // Generate the shader code for the color over lifetime gradient.
             // TODO - Move that to a pre-pass, not each frame!
-            let vertex_modifiers = if let Some(grad) = &asset.render_layout.lifetime_color_gradient
-            {
-                grad.to_shader_code()
-            } else {
-                String::new()
-            };
+            let mut vertex_modifiers =
+                if let Some(grad) = &asset.render_layout.lifetime_color_gradient {
+                    grad.to_shader_code()
+                } else {
+                    String::new()
+                };
+            if let Some(grad) = &asset.render_layout.size_color_gradient {
+                vertex_modifiers += &grad.to_shader_code();
+            }
+            trace!("vertex_modifiers={}", vertex_modifiers);
 
             // Configure the shader template, and make sure a corresponding shader asset exists
             let shader_source =
@@ -690,7 +823,7 @@ pub(crate) fn extract_effects(
             let shader = pipeline_registry.configure(&shader_source, &mut shaders);
 
             trace!(
-                "extracted: handle={:?} shader={:?} has_image={} position_code={}",
+                "extracted: handle={:?} shader={:?} has_image={} position_code={} force_field_code={}",
                 effect.handle,
                 shader,
                 if asset.render_layout.particle_texture.is_some() {
@@ -698,7 +831,8 @@ pub(crate) fn extract_effects(
                 } else {
                     "N"
                 },
-                position_code
+                position_code,
+                force_field_code,
             );
 
             extracted_effects.effects.insert(
@@ -709,6 +843,7 @@ pub(crate) fn extract_effects(
                     color: Color::RED, //effect.color,
                     transform: transform.compute_matrix(),
                     accel,
+                    force_field,
                     rect: Rect {
                         min: Vec2::ZERO,
                         max: Vec2::new(0.2, 0.2), // effect
@@ -723,45 +858,65 @@ pub(crate) fn extract_effects(
                         .map_or(HandleId::default::<Image>(), |handle| handle.id),
                     shader,
                     position_code,
+                    force_field_code,
                 },
             );
         }
     }
 }
 
+/// A single particle as stored in a GPU buffer.
 #[repr(C)]
 #[derive(Debug, Copy, Clone, Pod, Zeroable, AsStd430)]
 struct Particle {
+    /// Particle position in effect space (local or world).
     pub position: [f32; 3],
+    /// Current particle age in \[0:`lifetime`\].
     pub age: f32,
+    /// Particle velocity in effect space (local or world).
     pub velocity: [f32; 3],
-    pub pad: f32,
+    /// Total particle lifetime.
+    pub lifetime: f32,
 }
 
+/// A single vertex of a particle mesh as stored in a GPU buffer.
 #[repr(C)]
 #[derive(Copy, Clone, Pod, Zeroable)]
 struct ParticleVertex {
+    /// Vertex position.
     pub position: [f32; 3],
+    /// UV coordinates of vertex.
     pub uv: [f32; 2],
 }
 
-/// Global resource containing the GPU data to draw all the particle effects.
+/// Global resource containing the GPU data to draw all the particle effects in all views.
 ///
 /// The resource is populated by [`prepare_effects()`] with all the effects to render
-/// for the current frame, and consumed by [`queue_effects()`] to actually enqueue the
-/// drawning commands to draw those effects.
+/// for the current frame, for all views in the frame, and consumed by [`queue_effects()`]
+/// to actually enqueue the drawning commands to draw those effects.
 pub(crate) struct EffectsMeta {
+    /// Map from an entity with a [`ParticleEffect`] component attached to it, to the associated
+    /// effect slice allocated in an [`EffectCache`].
     entity_map: HashMap<Entity, EffectSlice>,
+    /// Global effect cache for all effects in use.
     effect_cache: EffectCache,
     /// Bind group for the camera view, containing the camera projection and other uniform
     /// values related to the camera.
     view_bind_group: Option<BindGroup>,
+    /// Bind group for the simulation parameters, like the current time and frame delta time.
     sim_params_bind_group: Option<BindGroup>,
+    /// Bind group for the particles buffer itself.
     particles_bind_group: Option<BindGroup>,
+    /// Bind group for the spawning parameters (number of particles to spawn this frame, ...).
     spawner_bind_group: Option<BindGroup>,
+    /// Bind group for the indirect buffer.
+    indirect_buffer_bind_group: Option<BindGroup>,
     sim_params_uniforms: UniformVec<SimParamsUniform>,
     spawner_buffer: BufferVec<SpawnerParams>,
-    /// TEMP
+    /// Unscaled vertices of the mesh of a single particle, generally a quad.
+    /// The mesh is later scaled during rendering by the "particle size".
+    // FIXME - This is a per-effect thing, unless we merge all meshes into a single buffer (makes
+    // sense) but in that case we need a vertex slice too to know which mesh to draw per effect.
     vertices: BufferVec<ParticleVertex>,
 }
 
@@ -784,6 +939,7 @@ impl EffectsMeta {
             sim_params_bind_group: None,
             particles_bind_group: None,
             spawner_bind_group: None,
+            indirect_buffer_bind_group: None,
             sim_params_uniforms: UniformVec::default(),
             spawner_buffer: BufferVec::new(BufferUsages::STORAGE),
             vertices,
@@ -835,6 +991,8 @@ pub struct EffectBatch {
     shader: Handle<Shader>,
     /// Update position code.
     position_code: String,
+    /// Update force field code.
+    force_field_code: String,
     /// Compute pipeline specialized for this batch.
     compute_pipeline: Option<ComputePipeline>,
 }
@@ -898,10 +1056,6 @@ pub(crate) fn prepare_effects(
         effects_meta.entity_map.insert(entity, slice);
     }
 
-    // Upload modified groups to GPU (either completely new group/buffer, or existing one with one or more
-    // newly added slices into it).
-    effects_meta.effect_cache.flush(&render_queue);
-
     // Deallocate GPU data for destroyed effect instances. This will automatically drop any group where
     // there is no more effect slice.
     for _entity in extracted_effects.removed_effect_entities.iter() {
@@ -945,6 +1099,8 @@ pub(crate) fn prepare_effects(
     let mut end = 0;
     let mut num_emitted = 0;
     let mut position_code = String::default();
+    let mut force_field_code = String::default();
+
     for (slice, extracted_effect) in effect_entity_list {
         let buffer_index = slice.group_index;
         let range = slice.slice;
@@ -989,6 +1145,7 @@ pub(crate) fn prepare_effects(
                         image_handle_id,
                         shader: shader.clone(),
                         position_code: position_code.clone(),
+                        force_field_code: force_field_code.clone(),
                         compute_pipeline: None,
                     },));
                     num_emitted += 1;
@@ -1017,6 +1174,16 @@ pub(crate) fn prepare_effects(
         position_code = extracted_effect.position_code.clone();
         trace!("position_code = {}", position_code);
 
+        force_field_code = extracted_effect.force_field_code.clone();
+        trace!("force_field_code = {}", force_field_code);
+
+        // extract the force field and turn it into a struct that is compliant with Std430,
+        // namely ForceFieldStd430
+        let mut extracted_force_field = [ForceFieldStd430::default(); FFNUM];
+        for (i, ff) in extracted_effect.force_field.iter().enumerate() {
+            extracted_force_field[i] = (*ff).into();
+        }
+
         // Prepare the spawner block for the current slice
         // FIXME - This is once per EFFECT/SLICE, not once per BATCH, so indeed this is spawner_BASE, and need an array of them in the compute shader!!!!!!!!!!!!!!
         let spawner_params = SpawnerParams {
@@ -1024,6 +1191,7 @@ pub(crate) fn prepare_effects(
             count: 0,
             origin: extracted_effect.transform.col(3).truncate(),
             accel: extracted_effect.accel,
+            force_field: extracted_force_field, // extracted_effect.force_field,
             seed: random::<u32>(),
             ..Default::default()
         };
@@ -1055,6 +1223,7 @@ pub(crate) fn prepare_effects(
                     image_handle_id,
                     shader: shader.clone(),
                     position_code: position_code.clone(),
+                    force_field_code: force_field_code.clone(),
                     compute_pipeline: None,
                 },));
                 num_emitted += 1;
@@ -1087,6 +1256,7 @@ pub(crate) fn prepare_effects(
             image_handle_id,
             shader: shader.clone(),
             position_code: position_code.clone(),
+            force_field_code: force_field_code.clone(),
             compute_pipeline: None,
         },));
         num_emitted += 1;
@@ -1111,9 +1281,13 @@ pub struct ImageBindGroups {
 #[derive(Default)]
 pub struct EffectBindGroups {
     /// Bind groups for each group index for compute shader.
-    values: HashMap<u32, BindGroup>,
+    update_particle_buffers: HashMap<u32, BindGroup>,
     /// Same for render shader.
-    render_values: HashMap<u32, BindGroup>,
+    render_particle_buffers: HashMap<u32, BindGroup>,
+    /// Bind groups for each indirect buffer associated with each particle buffer (update stage).
+    update_indirect_buffers: HashMap<u32, BindGroup>,
+    /// Bind groups for each indirect buffer associated with each particle buffer (render stage).
+    render_indirect_buffers: HashMap<u32, BindGroup>,
 }
 
 #[allow(clippy::too_many_arguments)]
@@ -1154,7 +1328,22 @@ pub(crate) fn queue_effects(
         }
     };
 
-    // Create the bind group for global the simulation parameters
+    if effects_meta.spawner_buffer.buffer().is_none() {
+        // No spawners are active
+        return;
+    }
+
+    // Create the bind group for the camera/view parameters
+    effects_meta.view_bind_group = Some(render_device.create_bind_group(&BindGroupDescriptor {
+        entries: &[BindGroupEntry {
+            binding: 0,
+            resource: view_binding,
+        }],
+        label: Some("particles_view_bind_group"),
+        layout: &render_pipeline.view_layout,
+    }));
+
+    // Create the bind group for the global simulation parameters
     effects_meta.sim_params_bind_group =
         Some(render_device.create_bind_group(&BindGroupDescriptor {
             entries: &[BindGroupEntry {
@@ -1165,7 +1354,7 @@ pub(crate) fn queue_effects(
             layout: &update_pipeline.sim_params_layout,
         }));
 
-    // Create the bind group for the spawner
+    // Create the bind group for the spawner parameters
     effects_meta.spawner_bind_group = Some(render_device.create_bind_group(&BindGroupDescriptor {
         entries: &[BindGroupEntry {
             binding: 0,
@@ -1182,51 +1371,88 @@ pub(crate) fn queue_effects(
     // Queue the update compute
     trace!("queue effects from cache...");
     for (buffer_index, buffer) in effects_meta.effect_cache.buffers().iter().enumerate() {
-        // Ensure all effect groups have a binding for the entire buffer of the group,
+        // Ensure all effect groups have a bind group for the entire buffer of the group,
         // since the update phase runs on an entire group/buffer at once, with all the
         // effect instances in it batched together.
         trace!("effect buffer_index=#{}", buffer_index);
         effect_bind_groups
-            .values
+            .update_particle_buffers
             .entry(buffer_index as u32)
             .or_insert_with(|| {
                 trace!(
-                    "Create new update bind group for buffer_index={}",
+                    "Create new particle update bind group for buffer_index={}",
                     buffer_index
                 );
                 render_device.create_bind_group(&BindGroupDescriptor {
                     entries: &[BindGroupEntry {
                         binding: 0,
-                        resource: buffer.binding(32768 * 32),
+                        resource: buffer.max_binding(),
                     }],
-                    label: Some(&format!(
-                        "particles_particles_bind_group_compute{}",
-                        buffer_index
-                    )),
+                    label: Some(&format!("vfx_particles_bind_group_update{}", buffer_index)),
                     layout: &update_pipeline.particles_buffer_layout,
                 })
             });
 
         effect_bind_groups
-            .render_values
+            .update_indirect_buffers
             .entry(buffer_index as u32)
             .or_insert_with(|| {
                 trace!(
-                    "Create new render bind group for buffer_index={}",
+                    "Create new indirect update buffer bind group for buffer_index={}",
                     buffer_index
                 );
                 render_device.create_bind_group(&BindGroupDescriptor {
                     entries: &[BindGroupEntry {
                         binding: 0,
-                        resource: buffer.binding(32768 * 32),
+                        resource: buffer.indirect_max_binding(),
                     }],
                     label: Some(&format!(
-                        "particles_particles_bind_group_render{}",
+                        "vfx_indirect_buffer_bind_group_update{}",
                         buffer_index
                     )),
+                    layout: &update_pipeline.indirect_buffer_layout,
+                })
+            });
+
+        // Same for the render pipeline, ensure all buffers have a bind group.
+        effect_bind_groups
+            .render_particle_buffers
+            .entry(buffer_index as u32)
+            .or_insert_with(|| {
+                trace!(
+                    "Create new particle render bind group for buffer_index={}",
+                    buffer_index
+                );
+                render_device.create_bind_group(&BindGroupDescriptor {
+                    entries: &[BindGroupEntry {
+                        binding: 0,
+                        resource: buffer.max_binding(),
+                    }],
+                    label: Some(&format!("vfx_particles_bind_group_render{}", buffer_index)),
                     layout: &render_pipeline.particles_buffer_layout,
                 })
             });
+
+        effect_bind_groups
+            .render_indirect_buffers
+            .entry(buffer_index as u32)
+            .or_insert_with(|| {
+                trace!(
+                    "Create new indirect render buffer bind group for buffer_index={}",
+                    buffer_index
+                );
+                render_device.create_bind_group(&BindGroupDescriptor {
+                    entries: &[BindGroupEntry {
+                        binding: 0,
+                        resource: buffer.indirect_max_binding(),
+                    }],
+                    label: Some(&format!(
+                        "vfx_indirect_buffer_bind_group_render{}",
+                        buffer_index
+                    )),
+                    layout: &update_pipeline.indirect_buffer_layout,
+                })
+            });
     }
 
     // Queue the update
@@ -1241,6 +1467,7 @@ pub(crate) fn queue_effects(
             &update_pipeline,
             ParticleUpdatePipelineKey {
                 position_code: batch.position_code.clone(),
+                force_field_code: batch.force_field_code.clone(),
             },
             &render_device,
         );
@@ -1249,17 +1476,12 @@ pub(crate) fn queue_effects(
         batch.compute_pipeline = Some(compute_pipeline.clone());
     }
 
-    // Queue the rendering
-    effects_meta.view_bind_group = Some(render_device.create_bind_group(&BindGroupDescriptor {
-        entries: &[BindGroupEntry {
-            binding: 0,
-            resource: view_binding,
-        }],
-        label: Some("particles_view_bind_group"),
-        layout: &render_pipeline.view_layout,
-    }));
+    // Loop over all cameras/views that need to render effects
     let draw_effects_function = draw_functions.read().get_id::<DrawEffects>().unwrap();
     for mut transparent_phase in views.iter_mut() {
+        trace!("Process new Transparent3d view");
+        // For each view, loop over all the effect batches to determine if the effect needs to be rendered
+        // for that view, and enqueue a view-dependent batch if so.
         for (entity, batch) in effect_batches.iter() {
             trace!(
                 "Process batch entity={:?} buffer_index={} spawner_base={} slice={:?}",
@@ -1337,38 +1559,6 @@ pub(crate) fn queue_effects(
                 distance: 0.0, // TODO ??????
             });
         }
-        // for (_buffer_index, _buffer) in effects_meta.effect_cache.buffers().iter().enumerate() {
-        //     let entity = *effects_meta.entity_map.keys().next().unwrap(); // TODO
-
-        //     //for (entity, batch) in effect_batches.iter_mut() {
-        //     // image_bind_groups
-        //     //     .values
-        //     //     .entry(batch.image.clone_weak())
-        //     //     .or_insert_with(|| {
-        //     //         let gpu_image = gpu_images.get(&batch.image).unwrap();
-        //     //         render_device.create_bind_group(&BindGroupDescriptor {
-        //     //             entries: &[
-        //     //                 BindGroupEntry {
-        //     //                     binding: 0,
-        //     //                     resource: BindingResource::TextureView(&gpu_image.texture_view),
-        //     //                 },
-        //     //                 BindGroupEntry {
-        //     //                     binding: 1,
-        //     //                     resource: BindingResource::Sampler(&gpu_image.sampler),
-        //     //                 },
-        //     //             ],
-        //     //             label: Some("particles_material_bind_group"),
-        //     //             layout: &render_pipeline.material_layout,
-        //     //         })
-        //     //     });
-
-        //     transparent_phase.add(Transparent3d {
-        //         draw_function: draw_effects_function,
-        //         pipeline: render_pipeline_id,
-        //         entity,
-        //         distance: 0.0, // TODO ??????
-        //     });
-        // }
     }
 }
 
@@ -1430,20 +1620,28 @@ impl Draw<Transparent3d> for DrawEffects {
             //let effect_group = &effects_meta.effect_cache.buffers()[0]; // TODO
 
             pass.set_render_pipeline(pipeline);
+
+            // Vertex buffer containing the particle model to draw. Generally a quad.
             pass.set_vertex_buffer(0, effects_meta.vertices.buffer().unwrap().slice(..));
+
+            // View properties (camera matrix, etc.)
             pass.set_bind_group(
                 0,
                 effects_meta.view_bind_group.as_ref().unwrap(),
                 &[view_uniform.offset],
             );
+
+            // Particles buffer
             pass.set_bind_group(
                 1,
                 effect_bind_groups
-                    .render_values
+                    .render_particle_buffers
                     .get(&effect_batch.buffer_index)
                     .unwrap(),
                 &[],
             );
+
+            // Particle texture
             if effect_batch
                 .layout_flags
                 .contains(LayoutFlags::PARTICLE_TEXTURE)
@@ -1462,14 +1660,16 @@ impl Draw<Transparent3d> for DrawEffects {
                 }
             }
 
-            let count = effect_batch.slice.end - effect_batch.slice.start;
+            let vertex_count = effects_meta.vertices.len() as u32;
+            let particle_count = effect_batch.slice.end - effect_batch.slice.start;
 
             trace!(
-                "Draw {} particles for batch from buffer #{}.",
-                count,
+                "Draw {} particles with {} vertices per particle for batch from buffer #{}.",
+                particle_count,
+                vertex_count,
                 effect_batch.buffer_index
             );
-            pass.draw(0..6, 0..count); // TODO vertex count
+            pass.draw(0..vertex_count, 0..particle_count);
         }
     }
 }
@@ -1569,25 +1769,32 @@ impl Node for ParticleUpdateNode {
                         //let effect_slice = effects_meta.get(&effect_entity);
                         // let effect_group =
                         //     &effects_meta.effect_cache.buffers()[batch.buffer_index as usize];
-                        let particles_bind_group =
-                            effect_bind_groups.values.get(&batch.buffer_index).unwrap();
+                        let particles_bind_group = effect_bind_groups
+                            .update_particle_buffers
+                            .get(&batch.buffer_index)
+                            .unwrap();
+
+                        let indirect_bind_group = effect_bind_groups
+                            .update_indirect_buffers
+                            .get(&batch.buffer_index)
+                            .unwrap();
 
                         let item_size = batch.item_size;
-                        let item_count = (batch.slice.end - batch.slice.start) / item_size;
-                        let workgroup_count = item_count / 64;
+                        let item_count = batch.slice.end - batch.slice.start;
+                        let workgroup_count = (item_count + 63) / 64;
 
                         let spawner_base = batch.spawner_base;
-                        let buffer_offset = batch.slice.start / item_size;
+                        let buffer_offset = batch.slice.start;
 
                         trace!(
-                        "record commands for pipeline of effect {:?} ({} items / {}B/item = {} workgroups) spawner_base={} buffer_offset={}...",
-                        batch.handle,
-                        item_count,
-                        item_size,
-                        workgroup_count,
-                        spawner_base,
-                        buffer_offset,
-                    );
+                            "record commands for pipeline of effect {:?} ({} items / {}B/item = {} workgroups) spawner_base={} buffer_offset={}...",
+                            batch.handle,
+                            item_count,
+                            item_size,
+                            workgroup_count,
+                            spawner_base,
+                            buffer_offset,
+                        );
 
                         // Setup compute pass
                         //compute_pass.set_pipeline(&effect_group.compute_pipeline);
@@ -1603,6 +1810,7 @@ impl Node for ParticleUpdateNode {
                             effects_meta.spawner_bind_group.as_ref().unwrap(),
                             &[spawner_base * SpawnerParams::std430_size_static() as u32],
                         );
+                        compute_pass.set_bind_group(3, indirect_bind_group, &[buffer_offset]);
                         compute_pass.dispatch(workgroup_count, 1, 1);
                         trace!("compute dispatched");
                     }
